from betfair_parser.spec.common import BaseMessage, EndpointType, Params, Request, Response
from betfair_parser.strenums import DocumentedEnum, doc


class ActionPerformed(DocumentedEnum):
    NONE = doc("No action was performed since last heartbeat, or this is the first heartbeat")
    CANCELLATION_REQUEST_SUBMITTED = doc("A request to cancel all unmatched bets was submitted since last heartbeat")
    ALL_BETS_CANCELLED = doc("All unmatched bets were cancelled since last heartbeat")
    SOME_BETS_NOT_CANCELLED = doc("Not all unmatched bets were cancelled since last heartbeat")
    CANCELLATION_REQUEST_ERROR = doc("There was an error requesting cancellation, no bets have been cancelled")
    CANCELLATION_STATUS_UNKNOWN = doc("There was no response from requesting cancellation, cancellation status unknown")


class HeartbeatReport(BaseMessage, frozen=True):
    """Response from heartbeat operation."""

    action_performed: ActionPerformed  # The action performed since your last heartbeat request.
    actual_timeout_seconds: int  # The actual timeout applied to your heartbeat request


class _HeartbeatParams(Params, frozen=True):
    # Maximum period in seconds that may elapse (without a subsequent heartbeat request),
    # before a cancellation request is automatically submitted on your behalf. The minimum
    # value is 10, the maximum value permitted is 300.
    preferred_timeout_seconds: int


class Heartbeat(Request, kw_only=True, frozen=True):
    """
    This heartbeat operation is provided to help customers have their positions managed automatically in the
    event of their API clients losing connectivity with the Betfair API. If a heartbeat request is not received
    within a prescribed time period, then Betfair will attempt to cancel all 'LIMIT' type bets for the given
    customer on the given exchange. There is no guarantee that this service will result in all bets being
    cancelled as there are a number of circumstances where bets are unable to be cancelled. Manual intervention
    is strongly advised in the event of a loss of connectivity to ensure that positions are correctly managed.
    If this service becomes unavailable for any reason, then your heartbeat will be unregistered automatically
    to avoid bets being inadvertently cancelled upon resumption of service. you should manage your position
    manually until the service is resumed. Heartbeat data may also be lost in the unlikely event of nodes
    failing within the cluster, which may result in your position not being managed until a subsequent heartbeat
    request is received.
    """

    endpoint_type = EndpointType.HEARTBEAT
    method = "HeartbeatAPING/v1.0/heartbeat"
    params: _HeartbeatParams
    return_type = Response[HeartbeatReport]



from typing import Union

from msgspec.json import Decoder

from betfair_parser.spec.betting.enums import MarketBettingType, MarketStatus, MarketTypeCode, RunnerStatus  # noqa
from betfair_parser.spec.streaming import enums, type_definitions
from betfair_parser.spec.streaming.enums import (
    ChangeType,
    LapseStatusReasonCode,
    MarketDataFilterFields,
    PriceLadderDefinitionType,
    SegmentType,
    StatusErrorCode,
)
from betfair_parser.spec.streaming.messages import (
    MCM,
    OCM,
    Authentication,
    Connection,
    Heartbeat,
    MarketSubscription,
    OrderSubscription,
    Status,
)
from betfair_parser.spec.streaming.type_definitions import (
    AvailableToBack,
    AvailableToLay,
    BestAvailableToBack,
    BestAvailableToLay,
    BestDisplayAvailableToBack,
    BestDisplayAvailableToLay,
    KeyLineDefinition,
    KeyLineSelection,
    MarketChange,
    MarketDataFilter,
    MarketDefinition,
    MarketFilter,
    MatchedOrder,
    Order,
    OrderFilter,
    OrderMarketChange,
    OrderRunnerChange,
    PriceLadderDefinition,
    RunnerChange,
    RunnerDefinition,
    StartingPriceBack,
    StartingPriceLay,
    StrategyMatchChange,
    Trade,
)


STREAM_REQUEST = Union[Authentication, MarketSubscription, OrderSubscription, Heartbeat]
STREAM_RESPONSE = Union[Connection, Status, MCM, OCM]
_STREAM_MESSAGES = Union[STREAM_RESPONSE, list[STREAM_RESPONSE], STREAM_REQUEST]
_STREAM_DECODER = Decoder(_STREAM_MESSAGES, strict=False)


def stream_decode(raw: bytes):
    return _STREAM_DECODER.decode(raw)


def stream_decode_lines(raw: bytes):
    return _STREAM_DECODER.decode_lines(raw)



from enum import Enum, _auto_null, auto  # noqa


class StrEnum(str, Enum):
    """Allow the `auto()` syntax to use the defined enum key as value.

    Unlike in python 3.11 StrEnum, the fieldnames are not lowered.

    class MyEnum(BaseEnum):
        FIELD = auto()

    >>> MyEnum.FIELD.value == "FIELD"
    True
    """

    def _generate_next_value_(key, start, count, last_values):
        return key

    def __repr__(self):
        return f"<{type(self).__name__}.{self.name}>"


class LowerStrEnum(StrEnum):
    """Like StrEnum, but have lowered values."""

    def _generate_next_value_(key, start, count, last_values):
        return key.lower()


class doc(auto):
    """Auto-generated enum field with docstring

    doc("docstring") replaces auto() for DocumentedEnums. A value
    can be set using doc(value=..., docstring=...), otherwise the
    same value as for auto() is calculated.

    This mechanism only works in combination with
    DocumentedEnum and messes up the values of an Enum otherwise.
    """

    def __init__(self, docstring=None, value=None):
        self._value = value
        self.__doc__ = docstring

    @property
    def value(self):
        # When value is accessed the first time by the enum internals, it
        # needs to return auto.value, in order to trigger the auto-generation
        # mechanism.
        if self._value is None:
            return _auto_null
        return self

    @value.setter
    def value(self, val):
        # value is set by the enum auto-generation mechanism
        self._value = val


class DocumentedEnum(Enum):
    """Enum with documentation strings.

    class DocEnum(DocumentedEnum):
        FIELD = doc("This is a docstring")
        FIELD2 = auto()

    >>> DocEnum.FIELD == "FIELD"
    True
    >>> DocEnum.FIELD.__doc__
    "This is a docstring"
    """

    def __new__(cls, val):
        member = object.__new__(cls)
        if isinstance(val, doc):
            member._value_ = val._value
            member.__doc__ = val.__doc__
        else:
            # also handle ordinary or auto() values
            member._value = val  # type: ignore
            member.__doc__ = None
        return member

    def _generate_next_value_(key, start, count, last_values):
        return key

    def __str__(self):
        if self.__doc__:
            return f"{self.name}: {self.__doc__}"
        return self.name

    def __repr__(self):
        return f"<{type(self).__name__}.{self.name}>"



from typing import Annotated, Literal
from urllib.parse import quote

import msgspec

from betfair_parser.exceptions import IdentityError, LoginImpossible
from betfair_parser.spec.common import BaseResponse, EndpointType, Params, Request, decode
from betfair_parser.strenums import DocumentedEnum, StrEnum, auto, doc


class IdentityRequest(Request, frozen=True):
    endpoint_type = EndpointType.IDENTITY

    def parse_response(self, response, raise_errors=True):
        resp = decode(response, type=self.return_type)
        if resp.is_error and raise_errors:
            exception = self.throws(str(resp.error), response=resp, request=self)
            raise exception
        return resp


class LoginStatus(StrEnum):
    SUCCESS = auto()
    LIMITED_ACCESS = auto()
    LOGIN_RESTRICTED = auto()
    FAIL = auto()


class LoginExceptionCode(DocumentedEnum):
    ACCOUNT_ALREADY_LOCKED = doc("the account is already locked")
    ACCOUNT_NOW_LOCKED = doc("the account was just locked")
    ACCOUNT_PENDING_PASSWORD_CHANGE = doc("The account must undergo password recovery to reactivate")
    ACTIONS_REQUIRED = doc("You must login to https://www.betfair.com to provide missing information.")
    AGENT_CLIENT_MASTER = doc("Agent Client Master")
    AGENT_CLIENT_MASTER_SUSPENDED = doc("Suspended Agent Client Master")
    AUTHORIZED_ONLY_FOR_DOMAIN_RO = doc("Attempting to login to the Betfair Romania with a non .ro account.")
    AUTHORIZED_ONLY_FOR_DOMAIN_SE = doc("Attempting to login to the Betfair Sweden with a non .se account.")
    BETTING_RESTRICTED_LOCATION = doc("the account is accessed from a location where betting is restricted")
    CERT_AUTH_REQUIRED = doc("Certificate required or certificate present but could not authenticate with it")
    CHANGE_PASSWORD_REQUIRED = doc("change password required")
    CLOSED = doc("the account is closed")
    DANISH_AUTHORIZATION_REQUIRED = doc("Danish authorization required")
    DENMARK_MIGRATION_REQUIRED = doc("Denmark migration required")
    DUPLICATE_CARDS = doc("duplicate cards")
    EMAIL_LOGIN_NOT_ALLOWED = doc("This account has not opted in to log in with the email")
    INPUT_VALIDATION_ERROR = doc("Input validation failed for not further specified reason")  # not listed
    INTERNAL_ERROR = doc("Not further specified internal error")  # not listed
    INTERNATIONAL_TERMS_ACCEPTANCE_REQUIRED = doc("The latest international terms and conditions must be accepted.")
    INVALID_CONNECTIVITY_TO_REGULATOR_DK = doc(
        "the DK regulator cannot be accessed due to some internal problems in the system behind or in "
        "at regulator; timeout cases included."
    )
    INVALID_CONNECTIVITY_TO_REGULATOR_IT = doc(
        "the IT regulator cannot be accessed due to some internal problems in the system behind or in at "
        "regulator; timeout cases included."
    )
    INVALID_USERNAME_OR_PASSWORD = doc("the username or password are invalid")
    ITALIAN_CONTRACT_ACCEPTANCE_REQUIRED = doc("The latest Italian contract version must be accepted")
    ITALIAN_PROFILING_ACCEPTANCE_REQUIRED = doc("You must login to the website to accept the new conditions")
    KYC_SUSPEND = doc("KYC suspended")
    MULTIPLE_USERS_WITH_SAME_CREDENTIAL = doc("There is more than one account with the same credential")
    NOT_AUTHORIZED_BY_REGULATOR_DK = doc(
        "the user identified by the given credentials is not authorized in the DK's jurisdictions due to the "
        "regulators' policies. Ex: the user for which this session should be created is not allowed to "
        "act(play, bet) in the DK's jurisdiction."
    )
    NOT_AUTHORIZED_BY_REGULATOR_IT = doc(
        "the user identified by the given credentials is not authorized in the IT's jurisdictions due to the "
        "regulators' policies. Ex: the user for which this session should be created is not allowed to "
        "act(play, bet) in the IT's jurisdiction."
    )
    PENDING_AUTH = doc("pending authentication")
    PERSONAL_MESSAGE_REQUIRED = doc("personal message required for the user")
    SECURITY_QUESTION_WRONG_3X = doc("the user has entered wrong the security answer 3 times")
    SECURITY_RESTRICTED_LOCATION = doc("the account is restricted due to security concerns")
    SELF_EXCLUDED = doc("the account has been self-excluded")
    SPAIN_MIGRATION_REQUIRED = doc("Spain migration required")
    SPANISH_TERMS_ACCEPTANCE_REQUIRED = doc("The latest Spanish terms and conditions version must be accepted")
    STRONG_AUTH_CODE_REQUIRED = doc("2 Step Authentication code is required.")
    SUSPENDED = doc("the account is suspended")
    SWEDEN_BANK_ID_VERIFICATION_REQUIRED = doc("Swedish bank id for Betfair.se not provided.")
    SWEDEN_NATIONAL_IDENTIFIER_REQUIRED = doc("Swedish National identifier for Betfair.se not provided.")
    TELBET_TERMS_CONDITIONS_NA = doc("Telbet terms and conditions rejected")
    TEMPORARY_BAN_TOO_MANY_REQUESTS = doc(
        "The limit for successful login requests per minute has been exceeded. "
        "New login attempts will be banned for 20 minutes"
    )
    TRADING_MASTER = doc("Trading Master Account")
    TRADING_MASTER_SUSPENDED = doc("Suspended Trading Master Account")

    # In case of a successful login, the error field is an empty string. To express this, either a
    # Union[LoginExceptionCode, Literal[""]] or a subclass of this enum, extending the fields, would
    # be the most explicit way to express this behaviour. Unfortunately, we can't have a union
    # of two string classes for unambiguous parsing with msgspec. And also unfortunately, python
    # enums can't be subclassed in a straight forward way. So let's add the successful status
    # codes here as well.
    NO_ERROR = doc(value="", docstring="No error occurred")  # for login
    SUCCESS = doc(value="SUCCESS", docstring="No error occurred, operation successful")  # for certlogin


LoginStatusCode = Annotated[LoginExceptionCode, msgspec.Meta(title="LoginStatusCode")]


class LoginResponse(BaseResponse, frozen=True):
    token: str  # session token
    product: str  # application key
    status: LoginStatus
    error: LoginExceptionCode

    @property
    def is_error(self):
        return self.status != LoginStatus.SUCCESS


class _LoginParams(Params, frozen=True):
    username: str  # The username to be used for the login

    # The password to be used for the login. For strong auth customers, this should
    # be their password with a two-factor auth code appended to the password string.
    password: str


class Login(IdentityRequest, kw_only=True, frozen=True):
    method = "login"
    params: _LoginParams
    return_type = LoginResponse  # type: ignore
    throws = LoginImpossible  # type: ignore

    @staticmethod
    def headers():
        return {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
            # X-Application to be set by the application
        }

    def body(self):
        return f"username={quote(self.params.username)}&password={quote(self.params.password)}"


class KeepAliveLogoutResponse(BaseResponse, frozen=True):
    token: str  # session token
    product: str  # application key
    status: Literal["SUCCESS", "FAIL"]
    error: Literal["INPUT_VALIDATION_ERROR", "INTERNAL_ERROR", "NO_SESSION", ""]

    @property
    def is_error(self):
        return self.status != "SUCCESS"


class KeepAlive(IdentityRequest, frozen=True):
    method = "keepAlive"
    return_type = KeepAliveLogoutResponse  # type: ignore
    throws = IdentityError  # type: ignore


class Logout(IdentityRequest, frozen=True):
    method = "logout"
    return_type = KeepAliveLogoutResponse  # type: ignore
    throws = IdentityError  # type: ignore


class CertLoginResponse(BaseResponse, frozen=True):
    session_token: str
    login_status: LoginStatusCode

    # Behave like LoginResponse
    @property
    def token(self):
        return self.session_token

    @property
    def status(self):
        return self.login_status

    @property
    def error(self):
        return self.login_status if self.is_error else None

    @property
    def is_error(self):
        return self.status != LoginExceptionCode.SUCCESS


class CertLogin(IdentityRequest, kw_only=True, frozen=True):
    # Subclassing Login would have been more obvious, but then mypy freaks out
    # due to a different return_type

    endpoint_type = EndpointType.IDENTITY_CERT
    method = "certlogin"
    params: _LoginParams
    return_type = CertLoginResponse  # type: ignore
    throws = LoginImpossible  # type: ignore

    headers = staticmethod(Login.headers)
    body = Login.body



from typing import Optional

from betfair_parser.exceptions import AccountAPINGException
from betfair_parser.spec.accounts.enums import IncludeItem, Wallet
from betfair_parser.spec.accounts.type_definitions import (
    AccountDetailsResponse,
    AccountFundsResponse,
    AccountStatementReport,
    CurrencyRate,
)
from betfair_parser.spec.common import EndpointType, Params, Request, Response, TimeRange


class AccountRequest(Request, frozen=True):
    endpoint_type = EndpointType.ACCOUNTS
    throws = AccountAPINGException  # type: ignore


class _GetAccountFundsParams(Params, frozen=True):
    wallet: Optional[Wallet] = None  # Name of the wallet in question. Global wallet is returned by default


class _GetAccountDetailsParams(Params, frozen=True):
    pass


class GetAccountFunds(AccountRequest, kw_only=True, frozen=True):
    """Returns the available to bet amount, exposure and commission information."""

    method = "AccountAPING/v1.0/getAccountFunds"
    params: _GetAccountFundsParams
    return_type = Response[AccountFundsResponse]


class GetAccountDetails(AccountRequest, kw_only=True, frozen=True):
    """Returns the details relating your account, including your discount rate and Betfair point balance."""

    method = "AccountAPING/v1.0/getAccountDetails"
    params: _GetAccountDetailsParams
    return_type = Response[AccountDetailsResponse]


class _GetAccountStatementParams(Params, frozen=True):
    locale: Optional[str] = None  # The language to be used where applicable. Defaults to account settings
    from_record: Optional[int] = None  # Specifies the first record that will be returned, defaults to 0
    record_count: Optional[int] = None  # Specifies the maximum number of records to be returned. Maximum 100

    # Return items with an itemDate within this date range. Both from and to date times are inclusive.
    # If from is not specified then the oldest available items will be in range. If to is not specified
    # then the latest items will be in range. nb. This itemDataRange is currently only applied when
    # includeItem is set to ALL or not specified, else items are NOT bound by itemDate.
    # Please note:  You can only retrieve account statement items for the last 90 days.
    item_date_range: Optional[TimeRange] = None
    include_item: Optional[IncludeItem] = None  # Which items to include, if not specified then defaults to ALL.
    wallet: Optional[Wallet] = None  # Which wallet to return statementItems for. Defaults to UK


class GetAccountStatement(AccountRequest, kw_only=True, frozen=True):
    """Return the account statement. Essentially a large list of your last profits and losses."""

    method = "AccountAPING/v1.0/getAccountStatement"
    params: _GetAccountStatementParams
    return_type = Response[AccountStatementReport]


class _ListCurrencyRatesParams(Params, frozen=True):
    from_currency: Optional[str] = None  # The currency from which the rates are computed. Only GBP for now.


class ListCurrencyRates(AccountRequest, kw_only=True, frozen=True):
    """Returns a list of currency rates based on given currency. Updates only once per hour."""

    method = "AccountAPING/v1.0/listCurrencyRates"
    params: _ListCurrencyRatesParams
    return_type = Response[list[CurrencyRate]]



from typing import Any, Generic, Literal, Optional, TypeVar, get_type_hints

import msgspec

from betfair_parser.exceptions import APIError, APINGException, JSONError
from betfair_parser.spec.common.enums import (
    AccountAPINGExceptionCode,
    APINGExceptionCode,
    EndpointType,
    JSONExceptionCode,
)


def decode(raw: bytes, type: Any = Any) -> Any:
    try:
        return msgspec.json.decode(raw, strict=False, type=type)
    except msgspec.DecodeError as e:
        raise JSONError(str(e)) from e


def encode(data: Any) -> bytes:
    try:
        return msgspec.json.encode(data)
    except msgspec.EncodeError as e:
        raise JSONError(str(e)) from e


class BaseMessage(msgspec.Struct, kw_only=True, forbid_unknown_fields=True, frozen=True, rename="camel"):
    @classmethod
    def parse(cls, raw):
        return decode(raw, type=cls)

    def validate(self):
        return bool(decode(encode(self), type=type(self)))

    def to_dict(self):
        return msgspec.structs.asdict(self)

    def replace(self, **kwargs):
        return msgspec.structs.replace(self, **kwargs)


class Params(BaseMessage, omit_defaults=True, frozen=True):
    """By default, don't send None and other default values in operation parameters."""


class BaseResponse(BaseMessage, frozen=True):
    """Base class for Response, ErrorResponse and identity responses."""

    @property
    def is_error(self):
        """True, if the response is some kind of error."""
        return False


class RPC(BaseMessage, frozen=True):
    jsonrpc: Literal["2.0"] = "2.0"
    id: int = 1


ResultType = TypeVar("ResultType")
ParamsType = TypeVar("ParamsType")
ErrorCode = TypeVar("ErrorCode")


class ExceptionDetails(BaseMessage, Generic[ErrorCode], kw_only=True, frozen=True):
    error_code: ErrorCode
    error_details: Optional[str] = None  # The stack trace of the error
    request_uuid: Optional[str] = msgspec.field(name="requestUUID", default=None)

    @property
    def code(self):
        return self.error_code


class ExceptionData(BaseMessage, frozen=True, rename=None):
    exception_name: Literal["APINGException", "AccountAPINGException"] = msgspec.field(name="exceptionname")
    APINGException: Optional[ExceptionDetails[APINGExceptionCode]] = None
    AccountAPINGException: Optional[ExceptionDetails[AccountAPINGExceptionCode]] = None

    @property
    def error(self):
        return getattr(self, self.exception_name)

    @property
    def code(self):
        return self.error.code


class RPCError(BaseMessage, frozen=True):
    code: int  # JSONExceptionCode or any other undocumented integer code, if it's an APINGException
    message: str  # Something like "DSC-0018", "AANGX-0011", ...
    data: Optional[ExceptionData] = None  # The interesting part

    @property
    def exception_code(self):
        """Return some form of ExceptionCode, whatever error we get."""
        return self.data.code if self.data else JSONExceptionCode(self.code)

    def __str__(self):
        return str(self.exception_code)


class Response(RPC, BaseResponse, Generic[ResultType], kw_only=True, frozen=True):
    """RPC response. Either an error or a result."""

    result: Optional[ResultType] = None
    error: Optional[RPCError] = None

    @property
    def is_error(self):
        return self.error is not None


class Request(RPC, Generic[ParamsType], kw_only=True, frozen=True):
    endpoint_type = EndpointType.NONE
    method: str = ""
    params: ParamsType = {}  # type: ignore
    return_type = Response  # not to be serialized, so no type definition
    throws = APINGException  # JSON error definition

    @classmethod
    def with_params(cls, request_id=1, **kwargs):
        params_cls = get_type_hints(cls)["params"]
        return cls(
            params=params_cls(**kwargs),
            method=cls.method,
            id=request_id,
        )

    @staticmethod
    def headers():
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            # X-Application to be set by the application
            # X-Authentication to be set by the application
        }

    def body(self):
        return encode(self)

    def parse_response(self, response, raise_errors=True):
        resp = decode(response, type=self.return_type)
        if resp.id != self.id:
            raise APIError(f"Response ID ({resp.id}) does not match Request ID ({self.id})")
        if resp.is_error:
            if raise_errors:
                exception = self.throws(
                    str(resp.error.exception_code),
                    response=resp,
                    request=self,
                    code=resp.error.exception_code,
                )
                raise exception
            return resp.error
        return resp.result



import asyncio
import itertools
import socket
import ssl
import urllib.parse
from typing import Optional

from betfair_parser.exceptions import StreamAuthenticationError, StreamError
from betfair_parser.spec.common import encode
from betfair_parser.spec.streaming import (
    STREAM_REQUEST,
    STREAM_RESPONSE,
    Authentication,
    Connection,
    Heartbeat,
    Status,
    stream_decode,
)


def create_ssl_socket(hostname, timeout: int = 15) -> ssl.SSLSocket:
    """Create ssl socket and set timeout."""
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    context.load_default_certs()
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    secure_sock = context.wrap_socket(s, server_hostname=hostname)
    secure_sock.settimeout(timeout)
    return secure_sock


class _BaseStream:
    _connection_id: Optional[str] = None

    def __init__(self, endpoint) -> None:
        self._endpoint = endpoint
        self._id_generator = itertools.count()

    @property
    def connection_id(self) -> str:
        return self._connection_id

    def unique_id(self) -> int:
        return next(self._id_generator)


class Stream(_BaseStream):
    _sock: Optional[ssl.SSLSocket] = None
    _io: Optional[socket.SocketIO] = None

    def connect(self) -> None:
        url = urllib.parse.urlparse(self._endpoint)
        self._sock = create_ssl_socket(url.hostname)
        self._sock.connect((url.hostname, url.port))
        self._io = socket.SocketIO(self._sock, "rwb")
        msg: Connection = self.receive()  # type: ignore
        self._connection_id = msg.connection_id

    def send(self, request: STREAM_REQUEST) -> None:
        if not self._io:
            raise StreamError("Stream is not connected")
        msg = encode(request) + b"\r\n"
        written_bytes = self._io.write(msg)
        if not len(msg) == written_bytes:
            raise StreamError(f"Incomplete request transfer: {written_bytes} of {len(msg)} bytes sent")

    def receive(self) -> STREAM_RESPONSE:
        if not self._io:
            raise StreamError("Stream is not connected")
        return stream_decode(self._io.readline())

    def close(self):
        if self._io is not None:
            self._io.close()
            self._io = None
        if self._sock is not None:
            try:
                self._sock.shutdown(socket.SHUT_RDWR)
            except OSError:
                pass
            self._sock.close()
            self._sock = None

    def authenticate(self, app_key: str, token: str) -> None:
        self.send(Authentication(id=self.unique_id(), app_key=app_key, session=token))
        msg: Status = self.receive()  # type: ignore
        if msg.is_error:
            raise StreamAuthenticationError(f"{msg.error_code.name}: {msg.error_message}")
        if msg.connection_closed:
            raise StreamAuthenticationError("Connection was closed by the server unexpectedly")

    def heartbeat(self) -> None:
        self.send(Heartbeat(id=self.unique_id()))

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()


class AsyncStream(_BaseStream):
    _reader: Optional[asyncio.StreamReader] = None
    _writer: Optional[asyncio.StreamWriter] = None

    async def connect(self) -> None:
        url = urllib.parse.urlparse(self._endpoint)
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.load_default_certs()
        self._reader, self._writer = await asyncio.open_connection(
            host=url.hostname,
            port=url.port,
            ssl=context,
            server_hostname=url.hostname,
            limit=1_000_000,
        )
        msg: Connection = await self.receive()  # type: ignore
        self._connection_id = msg.connection_id

    async def send(self, request: STREAM_REQUEST) -> None:
        if not self._writer:
            raise StreamError("Stream is not connected")
        msg = encode(request) + b"\r\n"
        self._writer.write(msg)
        await self._writer.drain()

    async def receive(self) -> STREAM_RESPONSE:
        if not self._reader:
            raise StreamError("Stream is not connected")
        data = await self._reader.readline()
        return stream_decode(data)

    async def close(self):
        self._reader = None  # does not need to be closed explicitly
        if self._writer:
            await self._writer.drain()

            # Abort the underlying transport (equivalent to socket.shutdown())
            transport = self._writer.transport
            if transport:
                transport.abort()

            # Close the writer (which should in turn close the underlying transport)
            self._writer.close()
            await self._writer.wait_closed()
        self._writer = None

    async def authenticate(self, app_key: str, token: str) -> None:
        await self.send(Authentication(id=self.unique_id(), app_key=app_key, session=token))
        msg: Status = await self.receive()  # type: ignore
        if msg.is_error:
            raise StreamAuthenticationError(f"{msg.error_code.name}: {msg.error_message}")
        if msg.connection_closed:
            raise StreamAuthenticationError("Connection was closed by the server unexpectedly")

    async def heartbeat(self) -> None:
        await self.send(Heartbeat(id=self.unique_id()))

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()



from betfair_parser.strenums import DocumentedEnum, StrEnum, auto, doc


class ChangeType(StrEnum):
    HEARTBEAT = auto()
    SUB_IMAGE = auto()
    RESUB_DELTA = auto()


class SegmentType(StrEnum):
    SEG_START = auto()
    SEG = auto()
    SEG_END = auto()


class PriceLadderDefinitionType(StrEnum):
    CLASSIC = auto()
    FINES = auto()
    LINE_RANGE = auto()


class StatusErrorCode(DocumentedEnum):
    # General errors not sent with id linking to specific request (as no request context)
    INVALID_INPUT = doc(
        "Failure code returned when an invalid input is provided (could not " "deserialize the message)"
    )
    TIMEOUT = doc("Failure code when a client times out (i.e. too slow sending data)")

    # Specific to authentication
    NO_APP_KEY = doc("Failure code returned when an application key is not found in the message")
    INVALID_APP_KEY = doc("Failure code returned when an invalid application key is received")
    NO_SESSION = doc("Failure code returned when a session token is not found in the message")
    INVALID_SESSION_INFORMATION = doc("Failure code returned when an invalid session token is received")
    NOT_AUTHORIZED = doc("Failure code returned when client is not authorized to perform the operation")
    MAX_CONNECTION_LIMIT_EXCEEDED = doc(
        "Failure code returned when a client tries to create more connections than allowed to"
    )
    TOO_MANY_REQUESTS = doc("Failure code returned when a client makes too many requests within a short time period")

    # Specific to subscription requests
    SUBSCRIPTION_LIMIT_EXCEEDED = doc(
        "Customer tried to subscribe to more markets than allowed to - set to 200 markets by default"
    )
    INVALID_CLOCK = doc(
        "Failure code returned when an invalid clock is provided on re-subscription "
        "(check initialClk / clk supplied)"
    )

    # General errors which may or may not be linked to specific request id
    UNEXPECTED_ERROR = doc("Failure code returned when an internal error occurred on the server")
    CONNECTION_FAILED = doc("Failure code used when the client / server connection is terminated")


class LapseStatusReasonCode(DocumentedEnum):
    """The reason that some portion or all of this order has been lapsed. Unset if no portion of the order is lapsed."""

    MKT_UNKNOWN = doc(
        "The market was unknown, presumably removed from the matcher (closed) between bet placement and matching."
    )
    MKT_INVALID = doc("The market was known about but in an invalid state.")
    RNR_UNKNOWN = doc("The runner was unknown, presumably removed between bet placement and matching.")
    TIME_ELAPSED = doc("The bet was waiting in the queue too long, so was lapsed for safety.")
    CURRENCY_UNKNOWN = doc("The bet's currency ID was not recognised by the matcher.")
    PRICE_INVALID = doc("The bet's price was invalid, e.g. outside the defined ladder for the market.")
    MKT_SUSPENDED = doc("The market was suspended at the time the bet came to be matched.")
    MKT_VERSION = doc(
        "The bet had a maximum market version set, and the market's version on matching was greater than this."
    )
    LINE_TARGET = doc("The bet was on a line market, but was requested targeting profit or payout.")
    LINE_SP = doc("The bet was on a line market, but was either a BSP bet directly or requested to PERSIST_TO_SP.")
    SP_IN_PLAY = doc("The bet was a BSP bet that had somehow come to be placed after turn-in-play.")
    SMALL_STAKE = doc("The bet's stake was worth less than half a penny in GBP.")
    PRICE_IMP_TOO_LARGE = doc(
        "When the bet came to be matched, the price available was better than its best-permitted price, "
        "suggesting a significant shift in the market, presumably due to a major incident, which may have "
        "rendered the bet unwanted."
    )


class MarketDataFilterFields(DocumentedEnum):
    EX_BEST_OFFERS_DISP = doc(
        "Best prices including Virtual Bets - depth is controlled by ladderLevels (1 to 10) - Please note: "
        "The virtual price stream is updated ~150 m/s after non-virtual prices. Virtual prices are calculated "
        "for all ladder levels."
    )
    EX_BEST_OFFERS = doc("Best prices not including Virtual Bets - depth is controlled by ladderLevels (1 to 10).")
    EX_ALL_OFFERS = doc("Full available to BACK/LAY ladder.")
    EX_TRADED = doc("Full traded ladder. This is the amount traded at any price on any selection in the market")
    EX_TRADED_VOL = doc("Market and runner level traded volume.")
    EX_LTP = doc("The 'Last Price Matched' on a selection.")
    EX_MARKET_DEF = doc("Send market definitions. To receive updates to any of the following fields")
    SP_TRADED = doc("Starting price ladder.")
    SP_PROJECTED = doc("Starting price projection prices. To receive any update to the Betfair SP Near and Far price.")



"""
Definition of the betfair streaming API messages as defined in:
- https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
- https://github.com/betfair/stream-api-sample-code/blob/master/ESASwaggerSchema.json
 """

from typing import List, Literal, Optional, Union

from betfair_parser.spec.common import BaseMessage
from betfair_parser.spec.streaming.enums import ChangeType, SegmentType, StatusErrorCode
from betfair_parser.spec.streaming.type_definitions import (
    MarketChange,
    MarketDataFilter,
    MarketFilter,
    OrderFilter,
    OrderMarketChange,
)


def first_lower(s: str) -> str:
    """Lower only the first character of a string."""
    return s[:1].lower() + s[1:]


class StreamRequest(BaseMessage, tag_field="op", tag=first_lower, omit_defaults=True, frozen=True):
    id: Optional[Union[int, str]] = None  # Client generated unique id to link request with response (like json rpc)


class StreamResponse(BaseMessage, tag_field="op", tag=str.lower, omit_defaults=True, frozen=True):
    id: Optional[Union[int, str]] = None  # Client generated unique id to link request with response (like json rpc)


class Authentication(StreamRequest, kw_only=True, frozen=True):
    """
    This message is the first message that the client must send on connecting to the server. You must
    be authenticated before any other request is processed.
    """

    app_key: str  # Application Key
    session: str  # Session Token generated by the API login


class _Subscription(StreamRequest, kw_only=True, frozen=True):
    """Common parent class for any Subscription request."""

    clk: Optional[str] = None  # Token value delta (received in MarketChangeMessage) for resuming a subscription
    conflate_ms: Optional[int] = None  # The conflation rate (looped back on initial image: bounds are 0 to 120000)
    heartbeat_ms: Optional[int] = None  # The heartbeat rate (looped back on initial image: bounds are 500 to 5000)
    initial_clk: Optional[str] = None  # Token value that should be passed to resume a subscription
    segmentation_enabled: bool = True  # allow server to send large sets of data in segments, instead of a single block


class MarketSubscription(_Subscription, kw_only=True, frozen=True):
    """
    This subscription type is used to receive price changes for one or more markets; your subscription
    criteria determine what you see. Limiting the amount of data that you consume will make your initial
    image much smaller (and faster) & suppress changes that are uninteresting to you.
    """

    market_filter: MarketFilter
    market_data_filter: MarketDataFilter


class OrderSubscription(_Subscription, kw_only=True, frozen=True):
    """This subscription type is used to receive order changes."""

    order_filter: OrderFilter  # Optional filter applied to order subscription


class Heartbeat(StreamRequest, frozen=True):
    """
    This is an explicit heartbeat request (in addition to the server heartbeat interval which is automatic).
    This functionality should not normally be necessary unless you need to keep a firewall open.
    """


class Connection(StreamResponse, kw_only=True, frozen=True):
    """This is received by the client when it successfully opens a connection to the server."""

    connection_id: str  # Unique identifier for support queries


class Status(StreamResponse, kw_only=True, frozen=True):
    """Every request receives a status response with a matching id."""

    connection_closed: bool
    connection_id: Optional[str] = None
    connections_available: Optional[int] = None  # The number of connections available for this account at this moment
    error_code: Optional[StatusErrorCode] = None
    error_message: Optional[str] = None  # Additional message in case of a failure
    status_code: Literal["SUCCESS", "FAILURE"]  # The status of the last request

    @property
    def is_error(self):
        return self.status_code != "SUCCESS"


class _ChangeMessage(StreamResponse, kw_only=True, frozen=True):
    """Common parent class for any ChangeMessage."""

    clk: Optional[str] = None  # Token value (non-None) should be stored for resuming in case of a disconnect
    conflate_ms: Optional[int] = None  # The conflation rate (may differ from that requested if subscription is delayed)
    ct: Optional[ChangeType] = None
    heartbeat_ms: Optional[int] = None  # Heartbeat rate (may differ from requested: bounds are 500 to 30000)
    initial_clk: Optional[str] = None  # Token value (non-None) should be stored for resuming in case of a disconnect
    pt: int  # Publish Time
    segment_type: Optional[SegmentType] = None  # denotes the beginning and end of a segmentation
    status: Optional[int] = None  # None if the stream is up-to-date and 503 if the services are experiencing latencies

    @property
    def is_heartbeat(self):
        return self.ct == ChangeType.HEARTBEAT

    @property
    def stream_unreliable(self):
        return self.status == 503

    @property
    def change_type(self):
        """Set to indicate the type of change (if None this is a delta)"""
        return self.ct

    @property
    def publish_time(self):
        """Time (in milliseconds since epoch) that the changes were generated"""
        return self.pt


class MCM(_ChangeMessage, kw_only=True, frozen=True):
    """This is the ChangeMessage stream of data Betfair sends back after subscribing to the market stream."""

    mc: Optional[List[MarketChange]] = None

    @property
    def market_changes(self):
        """The modifications to markets (will be None on a heartbeat)"""
        return self.mc


class OCM(_ChangeMessage, kw_only=True, frozen=True):
    """This is the ChangeMessage stream of data Betfair sends back after subscribing to the order stream."""

    oc: Optional[List[OrderMarketChange]] = None

    @property
    def order_market_changes(self):
        """The modifications to account's orders (will be None on a heartbeat)"""
        return self.oc
